/*
 * Common SPI-based screen support for the JZ4780.
 *
 * Copyright (C) 2018, 2020, 2021, 2023, 2024 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/devices/cpm-jz4780.h>
#include <l4/devices/dma-jz4780.h>
#include <l4/devices/gpio-jz4780.h>
#include <l4/devices/spi-hybrid.h>
#include <l4/devices/spi-jz4780.h>

#include <l4/devices/dma.h>
#include <l4/devices/memory.h>
#include "lcd-jz4780-spi-device.h"

#include <l4/re/c/dataspace.h>
#include <l4/re/env.h>
#include <l4/sys/cache.h>
#include <l4/sys/types.h>

#include <ipc/cap_alloc.h>
#include <ipc/mem_ipc.h>

#include <stdio.h>
#include <unistd.h>



/* Virtual addresses for the CPM, DMA, GPIO and SSI/SPI register blocks.

   NOTE: Should be able to use component abstractions for some of these
         eventually. */

static l4_addr_t cpm_virt_base = 0, cpm_virt_base_end = 0;
static l4_addr_t dma_virt_base = 0, dma_virt_base_end = 0;
static l4_addr_t gpio_virt_base = 0, gpio_virt_base_end = 0;
static l4_addr_t spi_virt_base = 0, spi_virt_base_end = 0;
static l4_addr_t spi_phys_base = 0, spi_phys_base_end = 0;

static l4_uint32_t dma_irq_start = 0, dma_irq_end = 0;

// CPM, DMA, GPIO, SSI/SPI and display device abstractions.

static Cpm_jz4780_chip *cpm_chip;
static Dma_jz4780_chip *dma_chip;
static Gpio_jz4780_chip *gpio_chip;
static Spi_jz4780_chip *spi_chip;

static Dma_channel *dma_channel;
static Spi_hybrid *spi_channel;
static Spi_channel *spi_jz4780_channel;



// Disable the display.

void
Lcd_jz4780_spi_device::disable()
{
  // NOTE: Need to support cancelling the transfer or disabling the peripheral.

  //disable_display();
}

// Configure the peripheral and enable the display.

void
Lcd_jz4780_spi_device::enable()
{
  // Set up the GPIO pins for the peripheral.
  // NOTE: Hard-coded pin usage!

  gpio_chip->setup(20, Hw::Gpio_chip::Output, 1);             // backlight
  gpio_chip->config_pad(21, Hw::Gpio_chip::Function_alt, 2);  // CE1
  gpio_chip->config_pad(28, Hw::Gpio_chip::Function_alt, 2);  // SCLK
  gpio_chip->config_pad(29, Hw::Gpio_chip::Function_alt, 2);  // MOSI
  gpio_chip->config_pad(30, Hw::Gpio_chip::Function_alt, 2);  // GPC

  // Initialise the screen.
  // NOTE: Specific to the ST7789.

  spi_channel->send_units(2, (const uint8_t []) {0x00, 0x01}, 2, 8);
  usleep(120000);
  spi_channel->send_units(2, (const uint8_t []) {0x00, 0x11}, 2, 8);
  usleep(5000);
  spi_channel->send_units(4, (const uint8_t []) {0x00, 0x36, 0x01, 0x00}, 2, 8);
  usleep(5000);
  spi_channel->send_units(10, (const uint8_t []) {0x00, 0x3a, 0x01, 0x05, 0x00, 0x13, 0x00, 0x29, 0x00, 0x21}, 2, 8);
  usleep(5000);
  spi_channel->send_units(22, (const uint8_t []) {0x00, 0x2a, 0x01, 0x00, 0x01, 0x00, 0x01, 0x00, 0x01, 0xef,
                                                  0x00, 0x2b, 0x01, 0x00, 0x01, 0x00, 0x01, 0x00, 0x01, 0xef,
                                                  0x00, 0x2c}, 2, 8);
  usleep(5000);

  for (l4_addr_t addr = get_framebuffer(); addr < get_framebuffer() + get_framebuffer_size(); addr += sizeof(uint32_t))
      *((uint32_t *) addr) = 0;

  l4_cache_clean_data(get_framebuffer(), get_framebuffer() + get_framebuffer_size());

  // Initialise the display transfer.
  // NOTE: Asserting the data signal.

  spi_channel->acquire_control(true);

  spi_jz4780_channel->transfer(_fb_region.vaddr, _fb_region.paddr,
                               get_framebuffer_size(), 2, 16,
                               _desc_region.vaddr, _desc_region.paddr);

  //enable_display();
}

// Set up memory addresses for the peripheral, initialising the framebuffer and
// descriptor members.

int
Lcd_jz4780_spi_device::_setup_memory()
{
  // Test for existing setup.

  if (_fb_region.vaddr)
    return 0;

  // Framebuffer and descriptor sizes.

  unsigned long fb_size = get_framebuffer_size(), desc_size = 32;

  // Allocate memory for the framebuffer and descriptors with 2**5 = 32 byte ==
  // 8 word alignment, sufficient for the descriptors.

  long err = get_dma_region(fb_size, 5, &_fb_region);

  if (err)
    return 1;

  err = get_dma_region(desc_size, 5, &_desc_region);

  if (err)
    return 1;

  return 0;
}

l4_size_t Lcd_jz4780_spi_device::get_framebuffer_size()
{
  // NOTE: Hard-coded size.

  return 240 * 240 * 2;
}

// Populate a view information structure.

void Lcd_jz4780_spi_device::get_view_info(l4re_video_view_info_t *view_info)
{
  // NOTE: Hard-coded properties.

  view_info->width                      = 240;
  view_info->height                     = 240;
  view_info->pixel_info.bytes_per_pixel = 2;
  view_info->bytes_per_line             = 480;

  view_info->pixel_info.r.shift = 11; view_info->pixel_info.r.size = 5;
  view_info->pixel_info.g.shift = 5;  view_info->pixel_info.g.size = 6;
  view_info->pixel_info.b.shift = 0;  view_info->pixel_info.b.size = 5;

  view_info->pixel_info.a.shift = 0;
  view_info->pixel_info.a.size  = 0;
}

// Access to peripheral memory.

static int setup_memory()
{
  if (get_memory("jz4780-cpm", &cpm_virt_base, &cpm_virt_base_end))
    return 1;

  if (get_memory("jz4780-dma", &dma_virt_base, &dma_virt_base_end))
    return 1;

  if (get_irq("jz4780-dma", &dma_irq_start, &dma_irq_end) < 0)
    return 1;

  if (get_memory("jz4780-gpio", &gpio_virt_base, &gpio_virt_base_end))
    return 1;

  if (get_memory_complete("jz4780-ssi", &spi_virt_base, &spi_virt_base_end,
                          &spi_phys_base, &spi_phys_base_end))
    return 1;

  // Initialise the abstractions.
  // NOTE: Using a preset GPIO configuration.

  int gpio_port = 1;

  cpm_chip = new Cpm_jz4780_chip(cpm_virt_base);

  dma_chip = new Dma_jz4780_chip(dma_virt_base, dma_virt_base_end, cpm_chip);

  dma_chip->enable();

  gpio_chip = new Gpio_jz4780_chip(gpio_virt_base, gpio_port);

  // Initialise the clocks for the SPI peripheral before obtaining the
  // peripheral abstraction.

  /* NOTE: Set a suitably high but arbitrary frequency to support the SPI bit
           clock. */

  cpm_chip->set_frequency(Clock_ssi, 100000000);

  spi_chip = new Spi_jz4780_chip(spi_phys_base, spi_virt_base,
                                 spi_virt_base_end, cpm_chip);

  // Obtain a DMA channel and an SPI channel for updating the display.
  // NOTE: Using preset channels, frequency and pin configuration.

  dma_channel = dma_chip->get_channel(12);

  spi_jz4780_channel = spi_chip->get_channel(1, dma_channel, 25000000);

  spi_channel = new Spi_hybrid(spi_jz4780_channel, gpio_chip, 30, 2);

  return 0;
}

// Device initialisation.

Lcd_jz4780_spi_device *Lcd_jz4780_spi_device::device = 0;

Lcd_device *Lcd_device::get_device()
{
  if (Lcd_jz4780_spi_device::device)
    return Lcd_jz4780_spi_device::device;

  if (setup_memory())
    return 0;

  // Initialise the common device.

  Lcd_jz4780_spi_device::device = new Lcd_jz4780_spi_device(NULL /* display_device */);

  return Lcd_jz4780_spi_device::device;
}
